/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <dirent.h> 
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>

/* 1. Type Definition                           */
/************************************************/
typedef signed char         int_08t;
typedef signed short        int_16t;
typedef signed int          int_32t;
typedef signed long long    int_64t;
typedef unsigned char      uint_08t;
typedef unsigned short     uint_16t;
typedef unsigned int       uint_32t;
typedef unsigned long long uint_64t;

typedef uint_32t INST_TYPE;
typedef uint_64t ADDR_TYPE;
typedef uint_64t DATA_TYPE;
typedef uint_64t data_t;


/* 2. Constant Definition                       */
/************************************************/
#define PROJ "SimCore"

#define DEF_FILE_NAME    "aout.txt"
#define BLOCK_SIZE       0x00002000 /* 8KB page */
#define BLOCK_TABLE_SIZE 0x00080000 /* page size*/
#define BLOCK_MASK       0x00001fff /* page mask*/
#define BLOCK_MASK_BIT   13         /* mask bit */
#define DATA_T_SIZE      8   /*  size of data_t */ 
#define DATA_T_BIT       3   /* s bit of data_t */

/** for AMASK op  and IMPLVER op **/
#define IMPLEMENTATION_VERSION 2   
#define CPU_FEATURE_MASK 0x0000000000000107ull 

#define FD_MAPPING_MAX 128      /* read/write  */
#define OSF_SIGCONTEXT_SIZE 568 /* in byte     */
#define OSF_FSTAT_SIZE       80 /* in byte     */
#define OSF_RUSAGE_SIZE      72 /* in byte     */
#define OPTION_MAX          30  /* max options */

#define BIT0    0x0000000000000001ull
#define BIT1    0x0000000000000002ull
#define BIT2    0x0000000000000004ull
#define BIT3    0x0000000000000008ull
#define BIT4    0x0000000000000010ull
#define BIT5    0x0000000000000020ull
#define BIT6    0x0000000000000040ull
#define BIT7    0x0000000000000080ull
#define BIT8    0x0000000000000100ull
#define BIT9    0x0000000000000200ull
#define BIT10   0x0000000000000400ull
#define BIT11   0x0000000000000800ull
#define BIT12   0x0000000000001000ull
#define BIT15   0x0000000000008000ull
#define BIT20   0x0000000000100000ull
#define BIT31   0x0000000080000000ull
#define BIT63   0x8000000000000000ull

#define EXTND8  0xffffffffffffff00ull
#define EXTND16 0xffffffffffff0000ull
#define EXTND21 0xffffffffffe00000ull
#define EXTND32 0xffffffff00000000ull
#define MASK08  0x00000000000000ffull
#define MASK16  0x000000000000ffffull
#define MASK21  0x00000000001fffffull
#define MASK32  0x00000000ffffffffull

#define ZAP0    0xffffffffffffff00ull
#define ZAP1    0xffffffffffff00ffull
#define ZAP2    0xffffffffff00ffffull
#define ZAP3    0xffffffff00ffffffull
#define ZAP4    0xffffff00ffffffffull
#define ZAP5    0xffff00ffffffffffull
#define ZAP6    0xff00ffffffffffffull
#define ZAP7    0x00ffffffffffffffull

/* 4. structure and class definition            */
/************************************************/

/************************************************/
class Env{
 public:
  class system_config      *sc;
  class evaluation_result  *e;
  class debug              *deb;
  class system_manager     *sys;
  class memory_system      *mem;
  class architecture_state *as;
};

/************************************************/
class main_memory{
  data_t *block_table[BLOCK_TABLE_SIZE];
  data_t *allocblock(data_t *);
 public:
  main_memory();
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
  int load_all(data_t*, data_t*);
};

/************************************************/
class architecture_state{
  Env *ev;
 public:
  data_t pc;        /* program counter      */
  data_t r[32];     /* general purpose regs */
  data_t f[32];     /* floating point  regs */
  uint_64t rr(int); /* register read        */
  architecture_state(Env*);
  architecture_state();
};

/************************************************/
class memory_system{
  Env *ev;
 public:
  class main_memory *mm;
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
  void ld_inst(data_t *, INST_TYPE *);
  void ld_nbyte(int, data_t *, data_t *);
  void st_nbyte(int, data_t *, data_t *);
  ~memory_system();
  memory_system(Env*);
};

/************************************************/
class system_manager{
  Env *ev;
  /* File Descriptor Mapping Table */
  DATA_TYPE fd_table[FD_MAPPING_MAX]; 
  /* Unique Data for rduniq & wruniq inst */
  DATA_TYPE uniq; 
 public:
  int running;    /* chip is running ? */
  void execute_pal(class instruction*);
  int translate_fd(DATA_TYPE);
  system_manager(Env*);
};


/* 5. function prototypes                       */
/************************************************/
/************************************************/
