/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include "define.h"

#define L_NAME "SimCore-Loader"
#define L_VER  "Version 1.3 2003-11-05"

/* sturctures to define the object code format  */
/*                         by Kenji Kise 1999.5 */
/************************************************/
struct DU_filehdr {
  unsigned short f_magic;
  unsigned short f_nscns; /* use this */
  int		 f_timdat;	
  long long	 f_symptr;	
  int		 f_nsyms;	
  unsigned short f_opthdr;	
  unsigned short f_flags;
};

/************************************************/
struct DU_aouthdr {
  short	magic;	  
  short	vstamp;	  
  short   bldrev;
  short   padcell;
  long long tsize;	/* use this */
  long long dsize;	/* use this */
  long long bsize;	
  long long entry;	/* use this */
  long long text_start;	
  long long data_start;	
  long long bbs_start;
  int	gprmask;	
  int	fprmask; 	
  long long	gp_value;
};

/************************************************/
struct DU_scnhdr {
  char        s_name[8]; /* use this */
  long long   s_paddr;   /* use this */
  long long   s_vaddr;
  long long   s_size;    /* use this */
  long long   s_scnptr;  /* use this */
  long long   s_relptr; 
  long long   s_lnnoptr;
  unsigned short s_nreloc;
  unsigned short s_nlnno; 
  int     	s_flags;  
};
/************************************************/


/************************************************/
void usage_loader(){
  char UsageMessage[] =  "\
 -s: use a 4-byte format instead of 8-byte. \n\
 -f[filiname]: specify the output filename. \n\
input executable must be alpha COFF format.\n\
";
  printf("\n");
  printf("Usage: %s [option] executable args\n",
	 L_NAME);
  printf("%s", UsageMessage); exit(0);
}

/************************************************/
architecture_state::architecture_state(){
  for(int i=0; i<32; i++) r[i]=0;
  for(int i=0; i<32; i++) f[i]=0;
}

/************************************************/
void format_check(char *p){
  FILE *fp;
  struct DU_filehdr hdr;
  struct DU_aouthdr aout;
  struct DU_scnhdr  st;

  if ((fp = fopen(p, "r")) == NULL) {
    fprintf(stderr, "Bad file name: %s\n", p);
    exit(0);
  }
  fread(&hdr,  sizeof(hdr), 1, fp);
  fread(&aout, sizeof(aout),1, fp);
  
  fprintf(stderr, "hdr.f_magic: %d,\n", 
	  hdr.f_magic);
  if(hdr.f_magic!=387){ /* elf 17791 */
    printf("%s : ",p);
    printf("This is not coff executable.\n");
    exit(0);
  }

  for(int i=0; i<hdr.f_nscns; i++){
    fread(&st, sizeof(st), 1, fp);
    fprintf(stderr, "name: %10s", st.s_name);
    fprintf(stderr, "   %09llx - %09llx\n", 
	    st.s_paddr, 
	    st.s_paddr + st.s_size);
    if(!strncmp(st.s_name, ".dynamic", 8)){
      printf("Using the shared library.\n");
      printf("opt -static for gcc.\n");
      printf("opt -non_shared for Digital.\n");
      exit(0);
    }
  }
  fclose(fp);
}

/************************************************/
void init_section(char *p, memory_system *mem, 
		  architecture_state *as){
  struct DU_filehdr hdr;
  struct DU_aouthdr aout;
  struct DU_scnhdr  st;
  
  FILE *fp = fopen(p, "r");
  fread(&hdr, sizeof(hdr), 1, fp);
  fread(&aout,sizeof(aout),1, fp);

  for(int i=0; i<hdr.f_nscns; i++){
    fread(&st, sizeof(st), 1, fp);
    if(!strcmp(st.s_name, ".sbss") || 
       !strcmp(st.s_name, ".bss") || 
       !strcmp(st.s_name, ".comment")) continue;
    
    int size = st.s_size / 4;
    int *data = new int[size];
    {
      FILE *f = fopen(p, "r");
      fseek(f, st.s_scnptr, SEEK_SET);
      fread(data, st.s_size, 1, f);
      fclose(f);
    }
    for(int j=0; j<size; j++){
      data_t dat = data[j];
      data_t adr = st.s_paddr + 4*j;
      mem->st_nbyte(4, &adr, &dat); /* store */
    }
    delete data;
  }
  fclose(fp);
  as->pc = aout.entry; /* Program Counter */
}

/************************************************/
#define BASE_STACK 0x11ff97000ull
#define BASE_ARGV  0x11ff98000ull
#define BASE_ARGP  0x11ff99000ull
void init_stack(int argc, char **argv, 
		memory_system *mem, 
		architecture_state *as){
  data_t dat, adr;
  adr = BASE_STACK;
  as->r[30] = adr; /* stack pointer */

  int argpc = 1;
  static char *argp[] ={"HOME=/"};

  /***** Error Checck *****/
  if(argc>15 || argpc>15){
    printf("Error: argc %d, argpc %d\n", 
	   argc, argpc);
    exit(1);
  }
  /***** argc *****/
  dat = argc;
  mem->st_nbyte(8, &adr, &dat);
  /***** argv *****/
  for(int i=0; i<argc; i++){
    adr = adr+8;
    dat = BASE_ARGV + 0x100*i;
    mem->st_nbyte(8, &adr, &dat);

    char *c = argv[i];
    for(unsigned int j=0; j<strlen(c); j++){
      data_t a = dat + j;
      data_t d = c[j];
      mem->st_nbyte(1, &a, &d);
    }
  }
  adr = adr+8;
  /***** argp *****/
  for(int i=0; i<argpc; i++){
    adr = adr+8;
    dat = BASE_ARGP + 0x100*i;
    mem->st_nbyte(8, &adr, &dat); /* argp */

    char *c = argp[i];
    for(unsigned int j=0; j<strlen(c); j++){
      data_t a = dat + j;
      data_t d = c[j];
      mem->st_nbyte(1, &a, &d);
    }
  }
}

/************************************************/
void create_image(architecture_state *as,
		  memory_system *mem,  char *f, 
		  int flag){
  FILE *fp;
  if(f!=NULL) fp = fopen(f, "w");
  else        fp = stdout;
  
  fprintf(fp, "/* SimCore 1.5 Image File */\n");

  fprintf(fp, "/*** Registers ***/\n");
  for(int i=0; i<32; i++)
    if(as->r[i])
      fprintf(fp, "/@reg %02d %016llx\n", 
              i, as->r[i]);
  fprintf(fp, "/@pc  32 %016llx\n", as->pc);
  
  fprintf(fp, "/*** Memory    ***/\n");
  data_t a, d;
  while(mem->mm->load_all(&a, &d)){
    if(flag==4){
      /***** little endian *****/
      fprintf(fp, "$%llx %08llx\n", a,
	      d & MASK32);
      fprintf(fp, "$%llx %08llx\n", a+4, 
	      (d>>32) & MASK32);
    }
    else{
      fprintf(fp, "@%llx %llx\n", a, d);
    }
  }
}

/** main function of SimCore Loader            **/
/************************************************/
int main(int argc, char **argv){
  fprintf(stderr, "%s %s\n", L_NAME, L_VER);

  if(argc==1) usage_loader();
  architecture_state as;
  Env e;
  memory_system mem(&e);

  int argc_n = argc-1;
  char *p = argv[1]; /* program name */
  char *outfile = NULL;
  char **argv_n = &argv[1];
  int flag = 8;

  char *opt = argv[1];
  if(!strncmp(opt, "-s", 2)){
    fprintf(stderr,"Use a 4-byte format.\n");
    flag = 4;
    argc_n --;
    argv_n = &argv[2];
    p = argv[2];
  }
  if(!strncmp(opt, "-f", 2)){
    outfile = &opt[2];
    printf("filename: %s\n", outfile);
    fflush(stdout);
    argc_n --;
    argv_n = &argv[2];
    p = argv[2];
  }

  format_check(p);
  init_section(p, &mem, &as);
  init_stack(argc_n, argv_n, &mem, &as);
  create_image(&as, &mem, outfile, flag);
  return 0;
}
/************************************************/
